import re
from typing import Dict

CLASSNAME_PATTERN = r"[.][a-z0-9-.:>~ )(\]\[]+"
KEY_PATTERN = r"[a-z-]+"
VALUE_PATTERN = r"[a-z0-9# )(-/%]+"
KEY_VALUE_PATTERN = rf"({KEY_PATTERN}):\s+({VALUE_PATTERN});"

# KEY_VALUE = re.compile(r"([a-z-]+):\s+([a-z0-9# )(-/%]+);")
# RULE = re.compile(
# r"([.][a-z0-9-.:>~ )(\]\[]+)\s+{(\s+[a-z-]+:[a-z0-9# )(-/%]+;)+\s*}"
# )
KEY_VALUE = re.compile(KEY_VALUE_PATTERN)
RULE = re.compile(
    rf"({CLASSNAME_PATTERN})"
    + r"\s*{"
    + rf"(\s*{KEY_PATTERN}:\s+{VALUE_PATTERN};)+"
    + r"\s*}"
)
DECIMAL = re.compile(r"0([.][0-9]+)")
SPACES = re.compile(r"\s+")

Stylesheet = Dict[str, Dict[str, str]]


def rgb_to_html(rgb_rule: str) -> str:
    rgb_rule = (
        rgb_rule.strip()
        .removeprefix("rgb(")
        .removesuffix(";")
        .removesuffix(")")
    )
    rgb_rule = rgb_rule.replace(",", " ").replace("/", " ")
    rgb_rule = SPACES.sub(" ", rgb_rule)

    args = rgb_rule.split(" ")
    if len(args) == 4:
        last = args.pop().strip()
        if last.endswith("%"):
            opacity = float(last.removesuffix("%")) / 100.0
        else:
            opacity = float(last)
    else:
        opacity = 1.0

    if len(args) != 3:
        raise RuntimeError("Bad color:", args)

    rgb = []
    for c in args:
        if c.endswith("%"):
            code = round(float(c.removesuffix("%")) * 2.55)
        else:
            code = int(c)
        rgb.append(code)

    # rgba
    rgb.append(round(opacity * 255))
    return "#" + "".join(hex(c).removeprefix("0x").rjust(2, "0") for c in rgb)


def normalize_key(key: str) -> str:
    if ">" in key:
        key = key.split(">")[0].strip()
    return key


def normalize_color(value: str) -> str:
    pattern = r"rgb("
    i = value.find(pattern)
    while i >= 0:
        j = value.find(")", i) + 1
        try:
            value = value[:i] + rgb_to_html(value[i:j]) + value[j:]
        except BaseException:  # pylint: disable=broad-exception-caught
            # do not replace
            pass
        i = value.find(pattern, j)
    return value


def normalize_size(value: str) -> str:
    if value.startswith("0."):
        value = value[1:]
    if value.startswith("-0."):
        value = "-" + value[2:]
    if value in ["0px", "0em", "0rem"]:
        value = "0"
    return value


def normalize_space(value: str) -> str:
    return SPACES.sub(" ", value)


def parse_css(raw: str) -> Stylesheet:
    # remove lines
    raw = raw.strip().replace("\n", "")
    css: Dict[str, Dict[str, str]] = {}
    for m in RULE.finditer(raw):
        key = m.group(1).strip()
        key = normalize_key(key)
        css[key] = {}
        for k in KEY_VALUE.finditer(m.group(0)):
            prop, value = k.groups()

            value = normalize_space(value)
            value = normalize_size(value)
            value = normalize_color(value)
            css[key][prop] = value

    return css



from hashlib import sha256
from typing import Dict, Literal, Tuple

from bs4 import BeautifulSoup, Tag
from bs4.formatter import Formatter

from retailwind.parser import Stylesheet, parse_css
from retailwind.tailwind import CSS


def hash_rule(props: Dict[str, str]) -> str:
    buffer = " ".join(f"{k}:{props[k]}" for k in sorted(props.keys()))
    return sha256(buffer.encode()).hexdigest()[:10]


def reverse_index(style: Stylesheet, hash_as_key: bool = True):
    hm = {}
    for cl, rules in style.items():
        h = hash_rule(rules)
        if h in hm:
            pass
            # print("warning: key already exist", cl, rules, h)
        if hash_as_key:
            hm[h] = cl
        else:
            hm[cl] = h

    return hm


class Translator:
    pseudo_classes = ["hover", "focus", "focus-visible"]

    def __init__(
        self,
        compiled: str,
        tailwind: str = CSS,
        on_error: Literal["raise", "mark", "ignore"] = "mark",
    ) -> None:
        self.tailwind_css = parse_css(tailwind)
        self.compiled_css = parse_css(compiled)

        self.ht = reverse_index(self.tailwind_css)
        self.hc = reverse_index(self.compiled_css, False)

        self.on_error = on_error

    def __raw_single_key(
        self,
        key: str,
    ) -> str:
        cl = key
        pc = ""

        if not cl.startswith("."):
            cl = "." + key

        if cl not in self.hc:
            for pc in self.pseudo_classes:
                if f"{cl}:{pc}" in self.hc:
                    cl = f"{cl}:{pc}"
                    pc = f"{pc}:"
                    break

        if cl not in self.hc:
            if self.on_error == "raise":
                raise KeyError(f"unknown class {key}")
            elif self.on_error == "mark":
                return f"!{key}"
            elif self.on_error == "ignore":
                return key
        try:
            return pc + self.ht[self.hc[cl]].removeprefix(".")
        except KeyError as err:
            if self.on_error == "raise":
                raise err
            elif self.on_error == "mark":
                return f"?{key}"
            elif self.on_error == "ignore":
                return key

    def raw(
        self,
        key_or_class: str,
        *others: str,
    ) -> str | Tuple[str, ...]:
        """Translate input class(es)

        The translator may fail. In this case, three behaviors can be used:
        - raise: the program raises an exception
        - mark: classes not found in the input css are prepend with a '!' (kind of error, it probably means that the provided stylesheet is not complete).
                Classes not found in tailwind.css are prepend with a '?' (it means that some classes are missing in the reference stylesheet).
        - ignore: initial classes are kept
        """
        keys = key_or_class.strip().split(" ")
        out = " ".join(self.__raw_single_key(k) for k in keys if len(k) > 0)
        if len(others) > 0:
            return out, *self.raw(*others)
        return out

    def html(self, html_or_str_input: str) -> str:
        """Replace obfuscated classes with tailwind ones on a HTML fragment"""
        soup = BeautifulSoup(html_or_str_input, "html.parser")
        if not soup.find():
            return "".join(self.raw(soup.get_text()))

        def recursive_re(tag: Tag):
            if "class" in tag.attrs:
                tag.attrs["class"] = [self.raw(c) for c in tag.attrs["class"]]  # type: ignore
            for c in tag.children:
                if isinstance(c, Tag):
                    recursive_re(c)

        recursive_re(soup)
        return str(soup)

    def prettify(self, html: str) -> str:
        """ """
        soup = BeautifulSoup(html, "html.parser")
        return soup.prettify(formatter=Formatter(indent=4))



from textual import widgets
from textual.app import App
from textual.color import Color
from textual.containers import Horizontal
from textual.events import Click, Paste

from retailwind.translator import Translator

PRIMARY = Color(231, 146, 13)
# DARK = Color(39, 39, 42)
DARK = Color.parse("#3f3f46")


class TextArea(widgets.TextArea):
    """TextArea with custom style"""

    BINDINGS = [
        ("escape", "quit", "Leave app"),
    ]

    def on_mount(self):
        self.styles.padding = 0
        self.styles.border = ("round", PRIMARY.with_alpha(0.6))
        self.styles.scrollbar_size_vertical = 1
        self.styles.scrollbar_size_horizontal = 1
        self.styles.scrollbar_color = PRIMARY
        self.styles.scrollbar_background_active = "black"
        self.styles.scrollbar_background_hover = "black"
        self.styles.scrollbar_color_hover = PRIMARY.with_alpha(0.9)
        self.styles.scrollbar_color_active = PRIMARY.with_alpha(0.9)


class Footer(widgets.Footer):
    """Footer with custom style"""

    DEFAULT_CSS = (
        """
    $accent: #fbbf24;
    $accent-darken-1: #f59e0b;
    $accent-darken-2: #d97706;
    """
        + widgets.Footer.DEFAULT_CSS
    )

    def on_mount(self):
        self.styles.background = DARK


class ReTailwindApp(App):
    """Interactive app"""

    ENABLE_COMMAND_PALETTE = False
    BINDINGS = [
        ("f5", "prettify_input()", "Prettify input"),
        ("f8", "clear_input()", "Clear input"),
    ]

    input_area = TextArea("", language="html", id="input", theme="vscode_dark")
    output_area = TextArea(
        "",
        language="html",
        id="output",
        theme="vscode_dark",
    )

    def __init__(self, translator: Translator, *args, **kwargs):
        self.translator = translator
        super().__init__(*args, **kwargs)
        # self._disabled_messages = {Click}

    def compose(self):
        yield widgets.Header()
        yield Horizontal(self.input_area, self.output_area)
        yield Footer()

    def action_prettify_input(self):
        self.input_area.text = self.translator.prettify(self.input_area.text)
        self.update_output()

    def action_clear_input(self):
        self.input_area.text = ""
        self.update_output()

    def update_output(self):
        self.input_area.post_message(self.input_area.Changed(self.input_area))

    def on_mount(self):
        self.title = "retailwind"
        self.sub_title = "Reverse obfuscated tailwind classes"
        self.styles.background = Color(0, 0, 0, 0)
        # self.input_area.focus()
        self.output_area.styles.border = ("round", DARK)
        # self.capture_mouse(None)

    def on_click(self, event: Click):
        pass

    def on_paste(self, event: Paste):
        self.input_area.text = event.text

    def on_text_area_changed(self, msg: widgets.TextArea.Changed):
        area: widgets.TextArea = msg.text_area
        if area.id == "input":
            try:
                output = self.translator.html(area.text)
                self.output_area.text = output
            except BaseException:  # pylint: disable=broad-exception-caught
                pass



from typing import Optional

import click
import requests

from retailwind.interactive import ReTailwindApp
from retailwind.tailwind import CSS
from retailwind.translator import Translator


def get_css(src: str):
    """Open css file"""
    if src.startswith("http://") or src.startswith("https://"):
        response = requests.get(src, timeout=5)
        if response.status_code != 200:
            raise RuntimeError(response.content)
        return response.content.decode()
    return open(src, "r", encoding="utf-8").read()


@click.command()
@click.option(
    "-c",
    "--css",
    help="Stylesheet of the html page that contains the obfuscated classes",
    required=True,
)
@click.option(
    "-r",
    "--raw",
    help="Raw class(es) (or HTML fragment) to deobfuscate",
    type=str,
    default=None,
)
@click.option(
    "-t",
    "--tailwind-css",
    type=str,
    help="Tailwind stylesheet different from the builtin one",
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    help="Run in interactive mode",
)
def main(
    css: str = "",
    raw: Optional[str] = None,
    tailwind_css: Optional[str] = None,
    interactive: bool = False,
):
    """Command entrypoint"""
    compiled = get_css(css)

    tailwind = CSS
    if tailwind_css is not None:
        tailwind = get_css(tailwind_css)

    # build the translator
    translator = Translator(
        compiled=compiled,
        tailwind=tailwind,
        on_error="ignore",
    )

    # cli run
    if not interactive:
        assert isinstance(raw, str)
        print(translator.html(raw))
        return

    # interactive run
    ReTailwindApp(translator).run()


if __name__ == "__main__":
    main()
